"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = __importStar(require("fs"));
const path = __importStar(require("path"));
class EFileUtil {
    static copyFile(source, destination) {
        return new Promise((resolve, reject) => {
            const rd = fs.createReadStream(source);
            rd.on('error', (err) => {
                reject(err);
            });
            const wr = fs.createWriteStream(destination);
            wr.on('error', (err) => {
                reject(err);
            });
            wr.on('close', () => {
                resolve();
            });
            rd.pipe(wr);
        });
    }
    static copyFolder(source, destination) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                fs.mkdir(destination, { recursive: true }, (err) => __awaiter(this, void 0, void 0, function* () {
                    if (err) {
                        reject(err);
                        return;
                    }
                    try {
                        const files = yield fs.promises.readdir(source, { withFileTypes: true });
                        for (const file of files) {
                            const srcPath = path.join(source, file.name);
                            const destPath = path.join(destination, file.name);
                            if (file.isDirectory()) {
                                yield this.copyFolder(srcPath, destPath);
                            }
                            else {
                                yield this.copyFile(srcPath, destPath);
                            }
                        }
                        resolve();
                    }
                    catch (err) {
                        reject(err);
                    }
                }));
            });
        });
    }
    static deleteFileOrDirectory(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const stats = yield fs.promises.lstat(filePath);
                if (stats.isDirectory()) {
                    const files = yield fs.promises.readdir(filePath);
                    const promises = files.map((file) => EFileUtil.deleteFileOrDirectory(path.join(filePath, file)));
                    yield Promise.all(promises);
                    yield fs.promises.rmdir(filePath);
                }
                else {
                    yield fs.promises.unlink(filePath);
                }
            }
            catch (err) {
                throw err;
            }
        });
    }
    static readFile(filePath, nobuffer = 'utf8') {
        return new Promise((resolve, reject) => {
            fs.readFile(filePath, nobuffer, (err, data) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve(data);
                }
            });
        });
    }
    static readDirectory(directoryPath, fileCallback) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const files = yield fs.promises.readdir(directoryPath);
                const promises = files.map((file) => __awaiter(this, void 0, void 0, function* () {
                    const filePath = path.join(directoryPath, file);
                    const stats = yield fs.promises.stat(filePath);
                    if (stats.isDirectory()) {
                        yield EFileUtil.readDirectory(filePath, fileCallback);
                    }
                    else {
                        if (fileCallback) {
                            fileCallback(filePath);
                        }
                    }
                }));
                yield Promise.all(promises);
            }
            catch (err) {
                throw err;
            }
        });
    }
    static writeFile(filePath, content) {
        return __awaiter(this, void 0, void 0, function* () {
            const directoryPath = path.dirname(filePath);
            try {
                fs.mkdirSync(directoryPath, { recursive: true });
                function logObjectType(obj) {
                    const objectType = Object.prototype.toString.call(obj);
                    return objectType.match(/\[object (\w+)\]/)[1];
                }
                if (typeof content !== 'string') {
                    content = Buffer.from(content);
                }
                fs.writeFileSync(filePath, content, 'utf8');
            }
            catch (_a) {
                console.log('保存错误', filePath);
            }
        });
    }
}
exports.default = EFileUtil;
